import logging
from itertools import chain

from .exceptions import ParseError
from .response import Response
from .utils import ensure_bytes

HTTP_LINE_SEPARATOR = b'\r\n'

logger = logging.getLogger(__name__)


class Connection(object):
    def __init__(self, server, reader, writer):
        self.server = server

        self.reader = reader
        self.writer = writer

        self.response = None

    async def run(self):
        env = {}

        line_it = self._read_line()
        first_line = await line_it.__anext__()
        env.update(_parse_first_line(first_line))

        async for line in line_it:
            name, value = _parse_header(line)

            if name == b'Host':
                host, port = _parse_server(value)
                env['SERVER_NAME'] = host
                env['SERVER_PORT'] = port
            else:
                key = name.upper().replace(b'-', b'_')
                if key in (
                    b'CONTENT_LENGTH',
                    b'CONTENT_TYPE',
                ):
                    env[key] = value

        logger.debug('env: %s', env)
        self.response = Response()

        app = self.server.get_app()
        ret = app(env, self.start_response)

        await self._write_line(b'HTTP/1.1 %s' % self.response.status.encode())

        if len(ret) == 1 and isinstance(ret[0], (bytes, str)):
            body = ensure_bytes(ret[0])

        content_length = len(body)

        headers = [
            ('Content-Length', str(content_length)),
        ]
        for key, value in chain(
            headers, self.response.headers
        ):
            await self._write_line(
                b'%s: %s' % (
                    ensure_bytes(key),
                    ensure_bytes(value),
                )
            )
        await self._write_line(b'')
        await self.writer.drain()

        self.writer.write(body)
        await self.writer.drain()

    async def _read_line(self):
        while True:
            line = await self.reader.readuntil(HTTP_LINE_SEPARATOR)

            # remove separator
            line = line[:-len(HTTP_LINE_SEPARATOR)]

            if not line:
                break

            yield line

    async def _write_line(self, line):
        self.writer.write(line + HTTP_LINE_SEPARATOR)

    def start_response(self, status, headers):
        self.response.status = status
        self.response.headers = headers


def _parse_first_line(line):
    env = {}

    parts = line.split(b' ')
    if len(parts) != 3:
        raise ParseError('first line parts count error: %d', len(parts))

    method, uri, version = parts

    path, query = _split_uri(uri)

    env['REQUEST_METHOD'] = method
    env['SCRIPT_NAME'] = ''
    env['PATH_INFO'] = path

    if query:
        env['QUERY_STRING'] = query

    env['SERVER_PROTOCOL'] = version

    return env


def _split_uri(uri):
    if b'?' in uri:
        parts = uri.split(b'?', 1)
        if len(parts) == 2:
            return parts
        else:
            raise ParseError('parse uri error: %s' % uri)
    else:
        return uri, ''


def _parse_header(line):
    parts = line.split(b':', 1)
    if len(parts) == 2:
        return parts[0], parts[1].strip()
    else:
        raise ParseError('parse header error: %s' % line)


def _parse_server(server):
    parts = server.split(b':', 1)
    if len(parts) == 2:
        host, port = parts
        port = int(port)
    else:
        host = server
        port = 80

    return host, port



import logging

from .event_loop import EventLoop
from .connection import Connection
from .utils import import_object

logger = logging.getLogger(__name__)


class Server(object):
    def __init__(self, app_path, loop=None):
        self.app_path = app_path

        if loop is None:
            loop = EventLoop.get_event_loop()

        self.loop = loop

    def run(self, *argv, **kwargs):
        self.loop.create_task(
            self.startup(*argv, **kwargs)
        )

        try:
            self.loop.run_forever()
        except KeyboardInterrupt:
            pass

    async def startup(self, port):
        await self.loop.start_server(
            self._client_connected,
            port=port
        )
        logger.info('server started on port %s' % (port,))

    async def _client_connected(self, reader, writer):
        logger.info('connected!')
        cnct = Connection(self, reader, writer)
        await cnct.run()

    def get_app(self):
        return import_object(self.app_path)



from asyncio import get_event_loop, start_server


class EventLoop(object):
    def __init__(self, real_loop):
        self._real = real_loop

    @classmethod
    def get_event_loop(cls):
        real_loop = get_event_loop()
        return cls(real_loop)

    def create_task(self, coro):
        def on_task_done(fut):
            assert fut.done(), 'task is not done'

            err = fut.exception()
            if err:
                self._real.stop()
                raise err

        task = self._real.create_task(coro)
        task.add_done_callback(on_task_done)

        return task

    def start_server(self, connected, port):
        return start_server(
            connected, port=port, loop=self._real
        )

    # pass by methods
    def run_forever(self):
        return self._real.run_forever()



from importlib import import_module


def ensure_bytes(value):
    if isinstance(value, str):
        value = value.encode('utf-8')

    return value


def import_object(path):
    module_path, obj_name = path.split(':')

    module = import_module(module_path)
    return getattr(module, obj_name)



from setuptools import setup

with open("README.rst", "r") as fh:
    long_description = fh.read()

setup(
    name='necrophos-wsgi',
    version='0.0.2',
    description='wsgi-like server based on asyncio',
    long_description=long_description,
    author='Elephant Liu',
    author_email='lexdene@gmail.com',
    url='https://github.com/lexdene/necrophos-wsgi',
    license='GPLv3',
    packages=['necrophos_wsgi'],
    entry_points={
        'console_scripts': [
            'necrophos_wsgi = necrophos_wsgi.main:main'
        ]
    },
)



import logging.config
from argparse import ArgumentParser


def parse_args():
    parser = ArgumentParser(description='start necrophos server')

    parser.add_argument(
        '-p', '--port',
        type=int,
        help='listen to port',
        default=8089,
    )

    parser.add_argument(
        '--app',
        required=True,
        help='app object',
    )

    return parser.parse_args()


def setup_logging():
    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'verbose': {
                'format': (
                    '[%(asctime)s %(process)d %(name)s '
                    '%(module)s:%(lineno)d %(levelname)s] %(message)s'
                )
            },
            'simple': {
                'format': '%(levelname)s %(message)s'
            },
        },
        'handlers': {
            'console': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'formatter': 'verbose',
            },
        },
        'loggers': {
            '': {
                'handlers': ['console'],
                'propagate': True,
                'level': 'DEBUG',
            },
            'sniper': {
                'propagate': True,
                'level': 'DEBUG',
            },

            'sqlalchemy.engine': {
                'level': 'INFO',
            },
        },
    })


def main():
    from .server import Server
    args = parse_args()

    setup_logging()

    server = Server(args.app)
    return server.run(port=args.port)
